; Z88 CamelForth
; multi.s (Multiprogramming)

; Words that execute PAUSE:
; EKEY (every 10ms)
; MS (every 10 ms)
; GFILL/GSHADE/GPATTERN (every line)
; GELLIPSE/GCIRCLE (every "step")
; REFILL (every block or file read)


;Z #TASKS	( -- addr )		Variable containing # active tasks
    head NUMTASKS,6,"#TASKS",docon,0,0
    	defw usr_numtasks

;Z LINK		( -- addr )		Variable containing link to next task
    head LINK,4,"LINK",docode,0,linkn-linkf
.linkf	push bc
	push iy
	pop hl
	ld bc,usr_link
	add hl,bc
	ld b,h
	ld c,l
.linkn
	next

;Z TERM		( -- addr )		LINK of the terminal task
    head TERM,4,"TERM",docon,0,0
    	defw pertask_start

; Each task has space reserved for a data stack, return stack and the "per
; task" region of user variables. The terminal task just has swapout space
; for the data stack reserved: this is done the first time that a task is
; created.
; Layout of region is:
;     LINK:					per-task uservariables
;     LINK+pertask_size:			data stack bottom
;     LINK+pertask_size+DATA_STACK_SIZE		return stack bottom
; First byte after region at LINK+pertask_size+DATA_STACK_SIZE+RET_STACK_SIZE

;Z TASK:	( "name" -- )		Create a new task
;   CURDS @ RAM
;   usr_s0 @ 0=				if necessary allocate terminal swap spc
;   IF  DATA_STACK_SPACE ALLOT  HERE usr_s0 ! THEN
;   CREATE  DATA_STACK_SPACE RET_STACK_SPACE pertask_size + + ALLOT
;   CURDS ! ;
    head TASKCOLON,5,"TASK:",docolon,0,0
	defw CURDS,FETCH,RAM
	defw LIT,pertask_start+usr_s0,FETCH,ZEROEQUAL,QBRANCH,CTASK1
	defw LIT,DATA_STACK_SIZE,ALLOT
	defw HERE,LIT,pertask_start+usr_s0,STORE
.CTASK1	defw CREATE
	defw LIT,DATA_STACK_SIZE+RET_STACK_SIZE+pertask_size,ALLOT
	defw CURDS,STORE,EXIT

;Z PAUSE	( -- )			Switch to next task (DEFERred)
;   DEFER PAUSE
    head PAUSE,5,"PAUSE",dodefer,0,0
    	defw usr_pause


; Separate versions of EKEY are provided for single and multi-tasked
; environments; the single version uses OS_IN and therefore allows the Z88
; to time out and switch off as normal. The multi version uses OS_TIN and
; therefore never times out.

;Z SINGLE	( -- )			Disable multitasking
;   ['] NOOP IS PAUSE
;   ['] EKEYS IS EKEY ;
    head SINGLE,6,"SINGLE",docode,0,0
    	ld hl,NOOP
	ld (usr_pause),hl
	ld hl,EKEYS
	ld (usr_ekey),hl
	next

;Z MULTI	( -- )			Enable multitasking
;   ['] (PAUSE) IS PAUSE
;   ['] EKEYM IS EKEY ;
    head MULTI,5,"MULTI",docode,0,0
    	ld hl,BRPAUSE
	ld (usr_pause),hl
	ld hl,EKEYM
	ld (usr_ekey),hl
	next

; Unfortunately, OZ requires that the machine stack be located below $2000
; in segment 0 when making OZ or pacakge calls, so we must swap the data
; stack to a reserved space when we switch tasks, swapping the data stack
; of the next task in to segment 0. Generally speaking, the data stack is
; going to be far smaller than the 128 cells allowed, so we take the trouble
; to calculate the actual size we need to transfer.

;Z (PAUSE)				Switch to next task
    nohead BRPAUSE,docode
    	push bc				; save TOS
	push ix				; save RP
	push de				; save IP
	ld hl,0
	add hl,sp
	ld (iy+usr_ticksp),l		; store SP into task's 'SP
	ld (iy+usr_ticksp+1),h

	ex de,hl
	ld hl,(initsp)
	and a
	sbc hl,de
	ld b,h
	ld c,l				; BC=size of stack to save
	ld hl,(initsp)
	dec hl				; HL=initsp-1, top address to save
	ld e,(iy+usr_s0)
	ld d,(iy+usr_s0+1)
	dec de
	lddr				; copy stack from seg 0 into task space

	ld l,(iy+usr_link)
	ld h,(iy+usr_link+1)
	push hl
	pop iy				; set IY to UP of next task in chain
	ld l,(iy+usr_ticksp)
	ld h,(iy+usr_ticksp+1)
	ld sp,hl			; get new task's SP from 'SP

	ex de,hl
	ld hl,(initsp)
	and a
	sbc hl,de
	ld b,h
	ld c,l				; BC=size of stack to restore
	ld l,(iy+usr_s0)
	ld h,(iy+usr_s0+1)
	dec hl
	ld de,(initsp)
	dec de
	lddr				; copy stack from task space into seg 0

	pop de				; restore IP
	pop ix				; restore RP
	pop bc				; restore TOS
	innext

;Z AWAKE?	( task -- a|0 )		If active, gives address of LINK in
;   >R #TASKS @ LINK			previous task
;   BEGIN  OVER
;   WHILE  SWAP 1- OVER @ DUP R@ =
;          IF  R> 2DROP DROP EXIT  THEN
;          ROT DROP
;   REPEAT
;   R> 2DROP DROP ZERO ;
    head AWAKEQ,6,"AWAKE?",docolon,0,0
    	defw TOR,NUMTASKS,FETCH,LINK
.awake1	defw OVER,QBRANCH,awake2
	defw SWOP,ONEMINUS,OVER,FETCH,DUP,RFETCH,EQUAL
	defw QBRANCH,awake3
	defw RFROM,TWODROP,DROP,EXIT
.awake3 defw ROT,DROP
	defw BRANCH,awake1
.awake2	defw RFROM,TWODROP,DROP,ZERO,EXIT

;Z SLEEP	( task -- )		Remove task from chain unless TERM
;   DUP TERM <>
;   IF  DUP AWAKE? ?DUP
;       IF  OVER @		( task prev next )
;           SWAP !  -1 #TASKS +!
;       THEN
;   THEN  DROP ;
    head SLEEP,5,"SLEEP",docolon,0,0
    	defw DUP,TERM,NOTEQUAL,QBRANCH,sleep1
	defw DUP,AWAKEQ,QDUP,QBRANCH,sleep1
	defw OVER,FETCH
	defw SWOP,STORE,MINUSONE,NUMTASKS,PLUSSTORE
.sleep1	defw DROP,EXIT

;Z STOP		( -- )			Puts current task to sleep
;   LINK SLEEP (PAUSE) ;
    head STOP,4,"STOP",docolon,0,0
    	defw LINK,SLEEP,BRPAUSE,EXIT

;Z WAKE		( task -- )		Inserts task into chain
;   DUP AWAKE? 0=
;   IF  DUP DUP LINK DUP @ 	( new new new curr next )
;       ROT ! !
;       1 #TASKS +!
;   THEN DROP ;
    head WAKE,4,"WAKE",docolon,0,0
   	defw DUP,AWAKEQ,ZEROEQUAL,QBRANCH,wake1
	defw DUP,DUP,LINK,DUP,FETCH
	defw ROT,STORE,STORE
	defw ONE,NUMTASKS,PLUSSTORE
.wake1	defw DROP,EXIT

;Z PURGE	( -- )			Kill all processes and ret to TERM
;   LINK				Does NOT disable multitasking
;   BEGIN  @ DUP SLEEP #TASKS 1 =  UNTIL
;   DROP PAUSE ;
    head PURGE,5,"PURGE",docolon,0,0
    	defw LINK
.purge1	defw FETCH,DUP,SLEEP
	defw NUMTASKS,FETCH,ONE,EQUAL
	defw QBRANCH,purge1
	defw DROP,PAUSE,EXIT

; The top-level word for each non-TERM task is the deferred word
; RUN, which normally just ensures any THROWn exceptions are caught
; and the task gracefully stopped. Users could reimplement this to report
; exceptions, for example.

;Z RUN	( xt -- )			Top-level word in subsidiary tasks
;   DEFER RUN ;				
    head RUN,3,"RUN",dodefer,0,0
    	defw usr_runtask

;Z (RUN)	( xt -- )		Default action for RUN
;   CATCH DROP STOP ;
    head BRRUN,5,"(RUN)",docolon,0,0
    	defw CATCH,DROP,STOP,EXIT

;Z TASK!	( xt task -- )		Set up a task ready to WAKE
;   LINK OVER pertask_size CMOVE		( copy current user vars )
;   DUP pertask_size+DATA_STACK_SIZE+RET_STACK_SIZE +
;   DUP >R OVER usr_r0 + !			( set R0 )
;   DUP pertask_size+DATA_STACK_SIZE +
;   DUP >R OVER usr_lp + !			( set LP )
;   R@ OVER usr_s0 + !				( set S0 )
;   R> R> OVER 4 - !				( "push" RP=R0 2nd)
;   usr_runtask OVER 6 - !			( "push" IP=usr_runtask 1st)
;   ROT SWAP 2 - !				( "push" TOS=xt 3rd)
;   initsp @ 6 - SWAP usr_ticksp + ! ;		( store SP in 'SP )
    head TASKSTORE,5,"TASK!",docolon,0,0
	defw LINK,OVER,LIT,pertask_size,CMOVE
	defw DUP,LIT,pertask_size+DATA_STACK_SIZE+RET_STACK_SIZE,PLUS
	defw DUP,TOR,OVER,LIT,usr_r0,PLUS,STORE
	defw DUP,LIT,pertask_size+DATA_STACK_SIZE,PLUS
	defw DUP,TOR,OVER,LIT,usr_lp,PLUS,STORE
	defw RFETCH,OVER,LIT,usr_s0,PLUS,STORE
	defw RFROM,RFROM,OVER,LIT,4,MINUS,STORE
	defw LIT,usr_runtask,OVER,LIT,6,MINUS,STORE
	defw ROT,SWOP,LIT,2,MINUS,STORE
	defw LIT,initsp,FETCH,LIT,6,MINUS,SWOP,LIT,usr_ticksp,PLUS,STORE
	defw EXIT
	
